(* Name     : Domain.thy - Domain construction
	      - an Isabelle theory for RALL -
   Author   : David von Oheimb, Technische Universitaet Muenchen
   Copyright: 1994

   created  : 20-Oct-94 direct product and power set
*)

Domain = Functionals +


consts

  (* Direct product *)
  Product1	:: "[('ab,'a)rel, ('ab,'b)rel] => bool"	("(Product1'(_,_'))")
  Product2	:: "[('ab,'a)rel, ('ab,'b)rel] => bool"	("(Product2'(_,_'))")
  Product3	:: "[('ab,'a)rel, ('ab,'b)rel] => bool"	("(Product3'(_,_'))")
  QuasiProd	:: "[('ab,'a)rel, ('ab,'b)rel] => bool"	("(QuasiProd'(_,_'))")
  Product	:: "[('ab,'a)rel, ('ab,'b)rel] => bool"	("(Product'(_,_'))")
  Pi		:: "('ab,'a)rel" 			("")
  Rho		:: "('ab,'b)rel" 			("")
  
  (* Power set *)
  Powerset1	:: "('s,'ps)rel => bool"
  Powerset2	:: "('s,'ps)rel => bool"
  Powerset	:: "('s,'ps)rel => bool"
  "@@Eps"	:: "('s,'ps)rel" 			("")


rules

  product1_def	"Product1(p,r) == Univalent(p) & Univalent(r)"
  product2_def	"Product2(p,r) == p^r=L"	   (* surjective *)
  product3_def	"Product3(p,r) == pp^  rr^ = I" (* compwise equal, total *)
  quasiprod_def	"QuasiProd(p,r)== Product1(p,r) & Product2(p,r)"
  product_def	"Product(p,r)  == QuasiProd(p,r) & Product3(p,r)"
  prod_pi_rho	"Product(,)"
  
  powerset1_def	"Powerset1(e)  == SyQ(e,e)  I"	 (* extensional equality *)
  powerset2_def	"Powerset2(e)  == R. Total(SyQ(R^,e))" (* comprehension *)
  powerset_def	"Powerset(e)   == Powerset1(e) & Powerset2(e)"
  powerset_eps	"Powerset()"


end
