(* Name     : Functionals.ML - theorems for Functionals.thy
   Author   : David von Oheimb, Technische Universitaet Muenchen
   Copyright: 1994

   created  : 05-Oct-94 residuals and Symmetric Quotients
   updated  : 04-Jan-95 adapted to new version of automatic proofs
*)

open Functionals;

val g  = g_  Functionals.thy;
val pg = pg_ Functionals.thy;


val Functionals1_ss = Special_ss addsimps [left_res_def, right_res_def,
					  sym_quot_def];
val fpre_tac    = pre_tac Functionals1_ss AtomRA_ss;
fun fpg     str = pg str [fpre_tac 1, REPEAT(rfast_tac 1)];


(**** Residuals **************************************************************)

val  left_res_mp = fpg "(SR)R  S";
val right_res_mp = fpg "R(RS)  S";

val  left_res_schr = fpg "Q  SR = QR  S";
val right_res_schr = fpg "R  QS = QR  S";

val res_equiv1 = fpg "(S^R^)^ = RS";
val res_equiv2 = fpg "(R^S^)= -R-S";

val  left_res_refl = fpg "Reflexive(RR)";
val right_res_refl = fpg "Reflexive(RR)";

val  left_res_trans = pg "(QR)(RS)  QS" [
				sstac left_res_schr 1,
				s_tac comp_assoc 1,
				mtac left_res_mp 1,
				rtac left_res_mp 1];
val right_res_trans = pg "(QR)(RS)  QS" [
				sstac right_res_schr 1,
				sstac comp_assoc 1,
				mtac right_res_mp 1,
				rtac right_res_mp 1];

val  left_res_preorder = pg "Preorder(RR)" [
				rewtac preorder_def,
				rtac conjI 1,
				rtac left_res_refl 1,
				rewtac transitive_def,
				rtac left_res_trans 1];
val right_res_preorder = pg "Preorder(RR)" [
				rewtac preorder_def,
				rtac conjI 1,
				rtac right_res_refl 1,
				rewtac transitive_def,
				rtac right_res_trans 1];

val  left_res_trans2 = pg "Q(RS)  (QR)S" [
				sstac left_res_schr 1,
				s_tac comp_assoc 1,
				mtac left_res_mp 1,
				rtac incl_refl 1];
val right_res_trans = pg "(QR)S  Q(RS)" [
				sstac right_res_schr 1,
				sstac comp_assoc 1,
				mtac right_res_mp 1,
				rtac incl_refl 1];

(**** Symmetric Quotient *****************************************************)

val sym_quot_equiv1 = fpg "-(R^-S)-(-R^S) = SyQ(R,S)";
val sym_quot_equiv2 = fpg "(RS)(R^S^) = SyQ(R,S)";
val sym_quot_equiv3 = fpg "(RS)(-R-S) = SyQ(R,S)";

val sym_quot_cmpl = fpg "SyQ(-R,-S) = SyQ(R,S)";
val conv_sym_quot = fpg "SyQ(R,S)^ = SyQ(S,R)";

(*SLOW*)
val sym_quot_isoton1 = fpg "SyQ(R,S)  SyQ(QR,QS)";
val function_sym_quot = pg "Function(F) ==> FSyQ(R,S) = SyQ(RF^,S)" [
				simp_tac Functionals1_ss 1,
				forward_tac [function_2a] 1,
				etac subst 1,
				forward_tac [function_2a] 1,
				s_tac comp_assoc 1,
				etac subst 1,
				forward_tac [function_2a] 1,
				s_tac comp_assoc 1,
				etac subst 1,
				rtac univalent_distr 1,
				rewtac function_def,
				fast_tac HOL_cs 1];

val sym_quot_incl = fpg "RSyQ(R,S)  S";
val sym_quot_eq = fpg "RSyQ(R,R) = R";

val sym_quot_pre_lemma = fpg "S-RSyQ(R,S) = O";
val sym_quot_pre = pg "RSyQ(R,S) = SLSyQ(R,S)" [
				s_tac cmpl_defL 1,
				sstac comp_distr_join2 1,
				sstac meet_distr 1,
				sstac sym_quot_pre_lemma 1,
				simp_tac Relation_ss 1,
				rtac incl_antisym 1,
				rtac meet_glb 1,
				rtac sym_quot_incl 1,
				rtac incl_refl 1,
				rtac meet_lb2 1];
val sym_quot_pre2 = pg "Total(SyQ(S,R)) ==> RSyQ(R,S) = S" [
				rewtac total_def,
				dtac (conv_equal RS iffD2) 1,
				asm_full_simp_tac (HOL_ss addsimps 
					[conv_sym_quot, conv_comp, conv_invol,
					 conv_L, sym_quot_pre]) 1,
				etac ssubst 1,
				simp_tac Lattice_ss 1];

val sym_quot_post = pg "SyQ(R,S)S^ = R^SyQ(R,S)L" [
				s_tac conv_equal 1,
				simp_tac(Relation_ss addsimps[conv_sym_quot])1,
				rtac sym_quot_pre 1];
(*SLOW 180sec*)
val sym_quot_red_lemma = fpg "SyQ(Q,R)SyQ(R,S)  SyQ(Q,S)"(* [
	pre_tac Functionals1_ss (* for efficiency *)
		(AtomRA_ss delsimps [atom_in_comp__] addsimps [atom_in_comp]) 1,
	REPEAT (r1fast_tac 1)]*);

val sym_quot_red1 = pg "SyQ(Q,R)SyQ(R,S) = SyQ(Q,S)SyQ(Q,R)L" [
				rtac incl_antisym 1,
				rtac meet_glb 1,
				rtac sym_quot_red_lemma 1,
				rtac comp_isoton2 1,
				rtac L_def 1,
				s_tac meet_commut 1,
				mon_tac dedekind_rule 1,
				mtac meet_lb1 1,
				simp_tac Lattice_ss 1,
				sstac conv_sym_quot 1,
				rtac comp_isoton2 1,
				rtac sym_quot_red_lemma 1];
val sym_quot_red2 = pg "SyQ(Q,R)SyQ(R,S) = SyQ(Q,S)LSyQ(R,S)" [
				rtac incl_antisym 1,
				rtac meet_glb 1,
				rtac sym_quot_red_lemma 1,
				rtac comp_isoton1 1,
				rtac L_def 1,
				s_tac meet_commut 1,
				mon_tac dedekind_rule 1,
				mtac meet_lb1 1,
				simp_tac Lattice_ss 1,
				sstac conv_sym_quot 1,
				rtac comp_isoton1 1,
				rtac sym_quot_red_lemma 1];

val sym_quot_refl = fpg "I  SyQ(R,R)";
val sym_quot_total = fpg "SyQ(R,R)L = L";
val sym_quot_trans = pg "SyQ(R,R)SyQ(R,S) = SyQ(R,S)" [
				rtac trans 1,
				rtac sym_quot_red1 1,
				sstac sym_quot_total 1,
				simp_tac Relation_ss 1];

val right_res_order = pg "Order(RR) = SyQ(R,R)  I" [
				rewtac order_def,
				simp_tac (HOL_ss addsimps 
					[right_res_preorder]) 1,
				simp_tac Functionals1_ss 1];


val Functionals_ss = Functionals1_ss addsimps [res_equiv1, res_equiv2,
				sym_quot_cmpl, conv_sym_quot,
				sym_quot_eq, sym_quot_total, sym_quot_trans];

