(* Name     : General.ML - General Tools for RALL
   Author   : David von Oheimb, Technische Universitaet Muenchen
   Copyright: 1994

   created  : 09-Nov-94 extracted from Lattice.ML
   updated  : 11-Nov-94 stable version of mtac, extracted from Relation.ML
   updated  : 18-Nov-94 debugging of inst_thm
   updated  : 01-Dec-94 new versions of g and pg
   updated  : 03-Nov-95 adaption for new read_insts
   updated  : 23-Aug-96 adaption for Isabelle94-6
   updated  : 04-Feb-97 adaptions for new HOL version
*)

print_mode := ["symbols"];

open General;

infixr 0 y;
fun _ y t = by t;
val b=9999;

fun  g_ theory formula       = let val prems = (goal theory formula) in
				 case prems of [] => ()
					     | _  => by(cut_facts_tac prems 1)
			       end;
fun pg_ theory formula tacsf = prove_goal theory formula
				 (fn []    => tacsf
				   | prems => (cut_facts_tac prems 1)::tacsf);

val  g =  g_ General.thy;
val pg = pg_ General.thy;

val set_cs = set_cs addIs [equalityI]; (* #### was missing *)

val mb = pg "[| P --> Q; ~Q |] ==> ~P" [fast_tac HOL_cs 1];

local
val contr1 = prove_goal HOL.thy "[| P; ~ P |] ==> Q" (fn prems => [
	cut_facts_tac prems 1,
	contr_tac 1]);
val contr2 = prove_goal HOL.thy "[| ~ P; P |] ==> Q" (fn prems => [
	cut_facts_tac prems 1,
	contr_tac 1]);
in
fun swapped_eresolve_tac thms = 
	(eresolve_tac (thms RL     [contr1] ) THEN' atac) INTLEAVE'
	(eresolve_tac (thms RLN (2,[contr2])) THEN' atac)
fun safe_swapped_eresolve_tac thms = 
	(ematch_tac (thms RL     [contr1] ) THEN' eq_assume_tac) ORELSE'
	(ematch_tac (thms RLN (2,[contr2])) THEN' eq_assume_tac)
end;
infix 4 addSEs2' addEs2' addss2;
fun css addEs2'  thms =  curry (pair_upd1 (op addaltern)) (css addEs2 thms)
			 (swapped_eresolve_tac thms);
fun css addSEs2' thms = curry (pair_upd1 (op addSaltern)) 
			(curry (pair_upd1 (op addaltern)) (css addSEs2 thms)
			 (swapped_eresolve_tac thms))
			(safe_swapped_eresolve_tac thms);
fun cs addss2 ss = cs compSWrapper (fn t => CHANGED o (full_simp_tac ss)
					    ORELSE' t);

fun lean_simp_tac thms = full_simp_tac (HOL_basic_ss 
			setSolver (K (K no_tac)) addsimps thms);




(**** misc *******************************************************************)

(* if test_tac succeeds on the proof state, then IF test_tac tac1 tac2
   applies tac1 to the same state, and tac2 otherwise *)
fun IF cond_tf then_tf else_tf = fn state => 
	case Sequence.pull(cond_tf state) of
		   None => else_tf state
		 | _    => then_tf state;

fun IF_SUBGOAL i = COND (has_fewer_prems i) all_tac;

fun APPEND_ALL' tacs = foldr (op APPEND') (tacs, K no_tac);
fun DETERM' f i = DETERM (f i);

(* this tests whether (thm1 RS thm2) would succeed *)
fun could_resolve thm1 thm2 = has_fewer_prems 0 (thm1 RS thm2) orelse true
			      handle THM _ => false;

(* this destruction tactic acts like dresolve_tac but 
   additionally solves the second premise of the rules by assumption *)
fun daresolve_tac thms = (dresolve_tac thms) THEN' assume_tac;
fun datac thm = daresolve_tac [thm];

fun ratac  thm  = rtac thm THEN' atac;
fun ra2tac thm i = rtac thm i THEN atac (i+1);

fun eatac  thm  = etac thm THEN' atac;

(**** premise manipulation ***************************************************)

(* double_prem_tac i duplicates the first premise of subgoal i *)
val double_prem = pg "P ==> P&P" [fast_tac HOL_cs 1];
val double_prem_tac = dtac double_prem THEN' etac conjE;

val swap2 = prove_goal HOL.thy "[| P; ~Q ==> ~P |] ==> Q" (fn prems => [
				cut_facts_tac prems 1,
				rtac swap 1,
				etac (hd(tl prems)) 2,
				fast_tac HOL_cs 1]) ;

fun det_rot_tac k = DETERM o (rotate_tac k);

fun kill_prem_tac n = DETERM o (rotate_tac n THEN' etac thin_rl);

local
  val mp2 = pg "[| P; P --> Q |] ==> Q" [fast_tac HOL_cs 1];
  val impI2 = prove_goal HOL.thy "[| R; (P&R ==> Q) |] ==> P --> Q" (fn prems=>[
                                cut_facts_tac prems 1,
                                rtac impI 1,
                                rtac (hd(tl(prems))) 1,
                                fast_tac HOL_cs 1]);
in fun APPLY_LIFT tf =     EVERY'[ REPEAT_DETERM o (etac mp2),
                                tf,
                                REPEAT_DETERM o (etac impI2 ORELSE' rtac impI),
                                REPEAT_DETERM o (etac conjE)];
end;

(**** substitution ***********************************************************)

(* s_tac equality_thm i substitutes the right hand side of the equality_thm
   in subgoal i, and sstac equality_thm i is the same for the left hand side *)
fun  s_tac thm = CHANGED o (rtac (thm RS  subst));
fun sstac thm = CHANGED o (rtac (thm RS ssubst));
fun rewr_tac thm = DETERM o (APPLY_LIFT (sstac thm));

val mp2 = pg "[| P; P --> Q |] ==> Q" [fast_tac HOL_cs 1];
val haz_ssubst_tac = EVERY'[ REPEAT_DETERM o (etac mp2), rtac impI, 
			     etac ssubst, strip_tac];

(* these do not work properly, see below 
fun inst_thm insts thm = let val {prop,sign,...} = rep_thm(topthm()) in
			read_instantiate_sg sign insts thm end;
fun inst_thm insts thm i = lift_inst_rule (topthm(), i, insts, thm);
*)

(* this version of read_instantiate takes the types of variables in the
   instantiated expressions from the variables of the given proof state *)
fun inst_thm state insts thm = let val {prop,sign,...} = rep_thm(state)
			     val (t1,s1) = types_sorts thm
  			     val (t2,s2) = types_sorts (state)
			     fun typ(a,i) = case t1(a,i) of
						Some(t) => Some(t)
					      | None    => t2(a,i)
			     fun sort(a,i) = case s1(a,i) of
						Some(s) => Some(s)
					      | None    => s2(a,i)
			     val ts = (typ,sort) in
			instantiate (read_insts sign ts ts [] insts) thm end;
(* this version of res_inst_tac instantiates thm2 using the types of variables
   in the instantiated expressions from the variables of the proof state,
   and performs  rtac (thm1 RS <resulting thm>) i *)
fun res2_inst_tac thm1 insts thm2 i state = state |> 
				(rtac (thm1 RS (inst_thm state insts thm2)) i);

(* like (s)s_tac, but str gives the subterm to be substituted in the subgoal *)
fun  subst_tac str thm = CHANGED o (res2_inst_tac thm [("t",str)] subst);
fun ssubst_tac str thm = CHANGED o (res2_inst_tac thm [("t",str)]ssubst);

(**** reasoning by monotonicity **********************************************)

(* mtac__ is a generic kind of substitution tactic like s_tac/sstac, but allows
   estimations by the monotonicity of several operators in inequality goal *)
fun mtac__ (trans_rule,const_left,const_right) mon_rules tac_on_subgoal i = let
  val res_mon = resolve_tac mon_rules
  val some_free_side = rtac const_left ORELSE' rtac const_right
  fun mtf i = fn s => let
    fun try_mon i = res_mon i THEN mtf i
    in (IF (rtac const_left i)
	   (IF (rtac const_right i)
	       (* both sides free:*)no_tac
	       (tac_on_subgoal i))
	   (tac_on_subgoal i) APPEND 
	IF (some_free_side i) (try_mon i)
	   (rtac trans_rule i THEN (try_mon i APPEND try_mon (i+1)))) s
    end
in mtf i end;

(*
usage of mtac__:
val mtac tac_o_s i =
mtac__ ((transitivity rule for the inequality predicate involved),
	(some thm of the form "some_prems ==> Some_Const < Some_Variable"),
	(some thm of the form "some_prems ==> Some_Variable < Some_Const"))
	(list of rules for monontonicity or antitonicity of some operators,
		 all having exactly one premise for mon./iso. in one argument)
	tac_o_s i;
with tac_o_s = tactic of type fn : int -> tactic employing some inequality,
		e.g. rtac inequality_thm  or  atac

local
  val mon = pg "s_st_t ==> s_st_t" [atac 1];
in
  (* like mtac, but the left side of thm  is instantiated by str *)
  fun m1tac str thm = mtac (thm RS (inst_thm [("s_s",str)] mon));
  (* like mtac, but the right side of thm  is instantiated by str *)
  fun m2tac str thm = mtac (thm RS (inst_thm [("t_t",str)] mon));
  val matac = mtac_ Relation_ms atac;
end;
*)

(* TO HOL.ML *)
val thin_True_rl = read_instantiate_sg (sign_of HOL.thy) [("V","True")] thin_rl;
