(* Name     : Lattice.thy - Complete Atomic Boolean Lattice
	      - an Isabelle theory for RALL -
   Author   : David von Oheimb, Technische Universitaet Muenchen 
   Copyright: 1994

   created  : 27-May-94
   updated  : 30-May-93 renamings
   updated  : 08-Jun-94 AtomIn
   updated  : 20-Jun-94 Completeness
   updated  : 13-Jul-94 new operator syntax
   updated  : 07-Feb-97 adaptions for new HOL version
*)

Lattice = General +


classes

  lattice < term


types dummy_lattice			(* to prevent a warning *)
arities dummy_lattice :: lattice	(* concerning empty sorts *)


consts

  (* Null and Universal Element *)
  O	:: "'a::lattice"
  L	:: "'a::lattice"
  
  (* Operations *)
  Meet	:: "['a::lattice,'a] => 'a"		("(_/_)"   [70,71] 70)
  Join	:: "['a::lattice,'a] => 'a"		("(_/_)"   [65,66] 65)
  Cmpl	:: " 'a::lattice     => 'a"		("(-_)"     [95]    95)

  Incl	:: "['a::lattice,'a] => bool"		("(_ / _)" [55,56] 55)

  (* Bounds of Sets; this implies their existence, i.e. completeness *)
  Inf	:: "'a::lattice set  => 'a"
  Sup   :: "'a::lattice set  => 'a"

  (* Inclusion of an Atom *)
  Atomic:: " 'a::lattice     => bool"
  AtomIn:: "['a,'a::lattice] => bool"		("(__)"    [52,52] 51)

rules

  meet_assoc	"(xy)z = x(yz)"	(* orientation is significant *)
  join_assoc	"(xy)z = x(yz)"	(* because of simplifier      *)

  meet_commut	"yx = xy"		(* orientation is significant *)
  join_commut	"yx = xy"		(* because of simplifier      *)

  meet_absorb	"x(xy) = x"
  join_absorb	"x(xy) = x"

  incl_def	"x  y  ==  (xy = x)"

  meet_hdistr1	"x(yz)  (xy)(xz)"

  O_def		"O  x"
  L_def		"x  L"

  inf_lb	"xA. Inf A  x"
  inf_glb	"(xA. l  x) --> l  Inf A"

  sup_ub	"xA. x  Sup A"
  sup_lub	"(xA. x  u) --> Sup A  u"

  cmpl_defL	"x-x = L"
  cmpl_defO	"x-x = O"

  atom_def	"Atomic a == aO & (x. xO & xa --> x=a)"
  atom_in_def	"ax == Atomic a & ax"
  weak_atomic	"xO --> (a. ax)"

end
