(* Name     : Relation.tac2.ML - tactics for Relation.thy
   Author   : David von Oheimb, Technische Universitaet Muenchen
   Copyright: 1995

   created  : 31-Jan-95, extracted from Relation.ML
   updated  : 02-Feb-95 corrections in ident_tac
*)

fun to_front_tac termp i = let
  fun find_eq n [] = None
  |   find_eq n (t :: ts) = if termp t then Some n else find_eq (n + 1) ts;
  fun tacf s = case find_eq 0 (Logic.strip_assums_hyp (#3(dest_state (s, i))))
		of None   => no_tac
	         | Some 0 => all_tac
	         | Some n => DETERM (rotate_tac n i);
in (fn st => tacf st st) end;
fun is_eq t = case t of (Const ("Trueprop", _) $ 
			(Const("op =",_) $ _ $ _)) => true 
			| _ => false;

fun is_var (Var _) = true
|   is_var (Const("at",_) $ d) = is_var d
|   is_var (Const("i",_) $ d) = is_var d
|   is_var (Const("Conv",_) $ d) = is_var d
|   is_var (v $ d) = is_var v
|   is_var _       = false;
local
fun is_atominI t = case t of (Const ("Trueprop", _) $ 
			(Const("AtomIn",_) $ a0 $ Const ("I", _))) =>
								not (is_var a0)
			| _ => false;
val rot_atominI = to_front_tac is_atominI;
in
fun ident_tacf I_D_ = REPEAT_DETERM1 o EVERY'[
			rot_atominI,
			dtac I_D_, etac conjE,
			REPEAT_DETERM o (etac mp2), rtac impI, rtac impI, 
			etac ssubst, etac conjE, strip_tac];
end;
val ident_tac = ident_tacf (pg "jI ==> j = i j & j^=j & Atomic j" [
		fast_tac (HOL_cs addSDs [I_D] addSEs [sym]) 1]);


local
fun incids s i = map (fn (n,(t,h)) => (n,h)) (fst (list_incids s i))
fun incid_pairs is = let fun pair (n,aas) (m,bbs) _ = 
			     let fun unifies a _ = flatmap (fn b => fn _ =>
				if not (could_unify (a,b)) then []
				else [(n,m,numeq a aas, numeq b bbs,
				       del a aas @ del b bbs)  ]) bbs
			     in flatmap unifies aas end
			in flatmap (fn x=> fn xs => flatmap (pair x) xs) is end
fun find_del_second _  []      = (false,[])
|   find_del_second p2 (x::xs) = if p2 x then (true,xs) else 
		     let val (v,xs2) = find_del_second p2 xs in (v,x::xs2) end
fun find_del_pairs _   []      = (~1,~1,~1,~1,[])
|   find_del_pairs p12 (x::xs) = case find_del_second (p12 x) xs of
			       (false ,_  ) => x
			     | (true  ,xs2) => find_del_pairs p12 xs2
in
fun assoc_tacf catac ass_thm i = let
    fun tacf s = let
	val ip=incid_pairs (incids s i);
	val (m,n,k,l,_) = if length ip<60 (** dirty boundary because of
					      termination problems **)  then
	   find_del_pairs (fn (_,_,_,_,x) => fn (_,_,_,_,z) =>is_permut x z) ip
	   else (~1,~1,~1,~1,[]) 
	in if m < 0 then no_tac else EVERY'[
		det_rot_tac  m   , forward_tac[ass_thm(k,l)],
	        det_rot_tac (n-m), catac l, etac exE, etac conjE] i end
    in REPEAT_DETERM1 (fn st => tacf st st) end;
end;
local
val assoc_thm = fn 
  (0,0) => comp_atom_assoc00
| (0,1) => comp_atom_assoc01
| (0,2) => comp_atom_assoc02
| (1,0) => comp_atom_assoc10
| (1,1) => comp_atom_assoc11
| (1,2) => comp_atom_assoc12
| (2,0) => comp_atom_assoc20
| (2,1) => comp_atom_assoc21
| (2,2) => comp_atom_assoc22
| (_,_) => sys_error "assoc_tac"
in
val assoc_tac = assoc_tacf conv_atac assoc_thm;
end;

