(* Name     : Special.ML - theorems for Special.thy
   Author   : David von Oheimb, Technische Universitaet Muenchen
   Copyright: 1994

   created  : 03-Aug-94
   updated  : 12-Aug-94 atomized proofs for total_2, total_3 and univalent_2b
   updated  : 18-Aug-94 proofs for order relations
   updated  : 29-Aug-94 proofs for vectors
   updated  : 06-Sep-94 proofs for points
   updated  : 04-Jan-95 adapted to new version of automatic proofs
*)

open Special;

val g  = g_  Special.thy;
val pg = pg_ Special.thy;

val Special_ss = Relation_ss addsimps[univalent_def, total_def , function_def,
				  injective_def, surjective_def, bijective_def,
				  difunctional_def,
				  reflexive_def, transitive_def, preorder_def,
				  antisymmetric_def, order_def,
				  symmetric_def, equivalence_def,
				  vector_def, predicate_def, point_def,
				  coreflexive_def];


val spre_tac  = pre_tac Special_ss AtomRA_ss;
fun spg     s = pg s [spre_tac 1, REPEAT (FIRSTGOAL rfast_tac)];
fun sslowpg s = pg s [spre_tac 1, REPEAT (FIRSTGOAL rslow_tac)];

val pred_type_incl = pg "RLSL ==> RLSL" [
				s_tac L_is_vector 1,
				sstac comp_assoc 1,
				matac 1,
				s_tac comp_assoc 1,
				s_tac L_is_vector 1,
				rtac incl_refl 1];
val pred_type_equiv = pg "RL=SL ==> RL=SL" [
				s_tac L_is_vector 1,
				s_tac L_is_vector 1, (*3*back();*)
				sstac comp_assoc 1,
				sstac comp_assoc 1,
				etac subst 1,
				rtac refl 1];

(**** Univalence *************************************************************)

val univalent_2a = spg "Univalent(R) ==> R-S  -(RS)"; (*assoc_tac/ident*)
val univalent_2b' = spg "R-I  -(RI) ==> Univalent(R)";
val univalent_2 = pg "Univalent(R) = ((S::('b,'b)rel). R-S  -(RS))" [
				safe_tac HOL_cs,
				etac univalent_2a 1,
				etac allE 1,
				etac univalent_2b' 1];

local
val univalent_3a_lemma =pg "Univalent(R) ==> Inf({R^(RQ) |Q. QA})  Inf(A)"[
				spre_tac 1,
				rotate_tac 1 1, rfast_tac 1];
in
val univalent_3a = 
	pg "Univalent(R) --> (X.XA) -->  (RInf(A) = Inf({RS |S. SA}))" [
				safe_tac HOL_cs,
				rtac incl_antisym 1,
				 rtac comp_hcontinous2 1,
				mon2_tac comp_isoton2 1,
				 etac univalent_3a_lemma 2,
				mon2_tac comp_isoton2 1,
				 rtac comp_hcontinous22 2,


				pre_tac (Special_ss delsimps [comp_assoc])
					(AtomRA_ss 
			delsimps [atom_in_comp__] addsimps [atom_in_comp]) 1,
				rotate_tac 3 1,
				double_prem_tac 1, (*  x. xA --> *)
				rotate_tac ~1 1,
				etac allE 1,
				 datac mp 1,
				safe_tac HOL_cs,
				etac allE 1,
				dtac mp 1,
				 rfast_tac 1,
				REPEAT(resolve_tac [exI, conjI] 1),
				  atac 1,
				 ratac conjI 1,
				REPEAT(resolve_tac [exI, conjI] 1),
				rfast_tac 1,
				rfast_tac 1];
end;
(*
(*SLOW 60sec*)
val univalent_3_lemma = rpg "Inf({RS|S. S{A,B}}) = Inf({RA, RB})";
*)
val univalent_3_lemma = pg "Inf({RS|S. S{A,B}}) = Inf({RA, RB})" [
				res_inst_tac [("f","Inf")] arg_cong 1,
				fast_tac set_cs 1];
val univalent_distr = pg "Univalent(R) ==> R(AB) = RARB" [
				s_tac inf_is_meet 1,
				s_tac inf_is_meet 1,
				cut_facts_tac [univalent_3a] 1,
				step_tac HOL_cs 1,
				contr_tac 1,
				step_tac HOL_cs 1,
				etac ssubst 2,
				fast_tac set_cs 1,
				rtac univalent_3_lemma 1];
val univalent_3b = pg
"(A. (RInf(A) = Inf({RS |(S::('b,'b)rel). SA}))) --> Univalent(R)" [
				rewtac univalent_def,
				rtac impI 1,
				etac allE 1,
				rtac schroeder_1b 1,
				sstac incl_equivO 1,
				simp_tac Relation_ss 1,
				s_tac I_def2 1, (*7*back();*)
				s_tac inf_is_meet 1,
				s_tac univalent_3_lemma 1,
				etac subst 1,
				simp_tac Relation_ss 1];
(****WRONG GOAL
val univalent_3 = pg "Univalent(R) = \
  \ (A. (S. SA) --> (RInf(A) = Inf({RS |(S::('b,'b)rel). SA})))" [
				safe_tac HOL_cs,
				dtac (univalent_3a RS mp) 1,
				 fast_tac HOL_cs 1,
				rtac (univalent_3b RS mp) 1,
				rtac allI 1,
				case_tac "A={}" 1,
				 
				];
****)

val univalent_4a = spg "Univalent(R) ==> (QSR^)R = QRS";
val univalent_4b' = pg "Q. (QIR^)R = QRI ==> Univalent(R)" [
				rewrite_goals_tac [univalent_def, incl_def],
				etac allE 1,
				etac subst 1,
				full_simp_tac Relation_ss 1];
val univalent_4 = pg "Univalent(R) = ((S::('b,'b)rel) Q. (QSR^)R = QRS)"[
				safe_tac HOL_cs,
				 etac univalent_4a 1,
				etac allE 1,
				etac univalent_4b' 1];

local val I_in_L = read_instantiate_sg (sign_of thy) [("x","I")] L_def in
val univalent_bump = pg "Univalent(R) ==> (SRQ = (R^SQ & SLRL))" [
				rewtac univalent_def,
				safe_tac HOL_cs,
				rtac pred_type_incl 2,
				dtac pred_type_incl 3,
				matac 1,
				sstac comp_assoc 1,
				matac 1,
				simp_tac Relation_ss 1,
				matac 1,
				s_tac L_is_vector 1,
				sstac comp_assoc 1,
				mtac L_def 1, (*2*back();*)
				rtac incl_refl 1,
				matac 1,
				m2tac "R^S" meet_lb2 1,
				mtac meet_lb1 1,
				mon2_tac dedekind_rule 1,
				matac 1,
				mtac I_in_L 1,
				simp_tac Relation_ss 1] end;

val univalent_equal = pg "Univalent(R) ==> (QR & RLQL) = (Q=R)" [
				safe_tac HOL_cs,
				dtac pred_type_incl 1,
				rewtac univalent_def,
				ratac incl_antisym 1,
				s_tac I_def2 1,
				matac 1,
				matac 1,
				s_tac meet_L 1, (*1*back();*)
				mtac meet_lb1 1,
				mon2_tac dedekind_rule 1,
				matac 1,
				mtac L_def 1,
				sstac I_def2 1,
				REPEAT(simp_tac Relation_ss 1)];
val univalent_equal1 = pg "[| Univalent(R); QR; RLQL |] ==> Q=R" [
				dtac univalent_equal 1,
				fast_tac HOL_cs 1];
val univalent_equal2 = pg "[| Univalent(R^); QR; LRLQ |] ==> Q=R" [
				s_tac conv_equal 1,
				etac univalent_equal1 1,
				etac conv_isoton2 1,
				rtac conv_isoton1 1,
				simp_tac Relation_ss 1];


(**** Totality ***************************************************************)


(*SLOW 30sec*)
val total_2 = spg "Total(R) = IRR^";		 (* requires trace_trace_D *)
val total_3 = spg "Total(R) = (S. -(RS)  R-S)";


(**** Functions **************************************************************)

val function_2a = pg "Function(R) ==> R-S = -(RS)" [
				rewtac function_def,
				etac conjE 1,
				rtac incl_antisym 1,
				etac univalent_2a 1,
				dtac (total_3 RS iffD1) 1,
				fast_tac HOL_cs 1];
val function_2 = pg "Function(R) = ((S::('b,'b)rel). R-S = -(RS))" [
				rewtac function_def,
				sstac univalent_2 1,
				sstac total_3 1,
				simp_tac (HOL_ss addsimps [equal_equiv]) 1,
				fast_tac HOL_cs 1];

val function_bump = pg "Function(R) ==> SRQ = R^SQ" [
				rewtac function_def,
				etac conjE 1,
				dtac univalent_bump 1,
				safe_tac HOL_cs,
				contr_tac 1,
				contr_tac 1,
				contr_tac 1,
				atac 2,
				atac 2,
				asm_full_simp_tac Special_ss 1,
				etac swap 1,
				etac swap 1,
				rtac incl_trans 1,
				atac 2,
				rtac L_def 1];
(**** combinations ***********************************************************)

val univalent_and_surjective = pg "(Univalent(R) & Surjective(R)) = (R^R=I)" [
				rewrite_goals_tac [univalent_def,
						   surjective_def],
				sstac total_2 1,
				spre_tac 1];

val total_and_injective = pg"(Total(R) & Injective(R)) = (RR^=I)" [
				sstac total_2 1,
				rewrite_goals_tac[injective_def,univalent_def],
				spre_tac 1];


(**** Difunctionality ********************************************************)

val difunctional_2 = pg "Difunctional(R) = (RR^R=R)" [
				rewtac difunctional_def,
				rtac iffI 1,
				etac incl_antisym 1,
				etac incl_intro1 2,
				rpre_tac 1,
				rfast_tac 1];

val univ_difunct = spg "Univalent(R) ==> Difunctional(R)";
val inj_difunct = spg "Injective(R) ==> Difunctional(R)";
val conv_difunct = pg "Difunctional(R^) = Difunctional(R)" [
				rewtac difunctional_def,
				s_tac conv_isoton 1,
				simp_tac Relation_ss 1];


(**** Reflexivity ************************************************************)

(*SLOW 60sec*)
val reflexive_2a = spg "Reflexive(R) ==> -(SR)  -(SR)R^";
val reflexive_2b = pg"((S::('b,'b)rel). -(SR)  -(SR)R^) ==> Reflexive(R)"[
				rewtac reflexive_def,
				step_tac HOL_cs 1,
				s_tac I_def1 1,
				rtac (cmpl_antiton RS iffD2) 1,
				etac incl_trans 1,
				s_tac schroeder2 1,
				rtac incl_refl 1];
val reflexive_2 = pg "Reflexive(R) = ((S::('b,'b)rel). -(SR)  -(SR)R^)" [
				safe_tac HOL_cs,
				etac reflexive_2a 1,
				etac reflexive_2b 1];


(**** Transitivity ***********************************************************)

val transitive_2a = pg "Transitive(R) ==> -(SR)R^  -(SR)" [
(*####				pre_tac Special_ss set_ss 1,
				full_simp_tac AtomRA_ss 1,*)
				rewtac transitive_def,
				s_tac schroeder2 1,
				s_tac comp_assoc 1,
				matac 1,
				rtac incl_refl 1];
val transitive_2b=pg"((S::('b,'b)rel). -(SR)R^  -(SR)) ==> Transitive(R)"[
				rewtac transitive_def,
				step_tac HOL_cs 1,
				s_tac I_def1 1, (*2*back();*)
				etac (schroeder2 RS iffD2) 1];
val transitive_2 = pg "Transitive(R) = ((S::('b,'b)rel). -(SR)R^  -(SR))"[
				safe_tac HOL_cs,
				etac transitive_2a 1,
				etac transitive_2b 1];


(**** Preorder ***************************************************************)

val preorder_2 = pg "Preorder(R) = ((S::('b,'b)rel). -(SR)R^ = -(SR))" [
				rewtac preorder_def,
				sstac reflexive_2 1,
				sstac transitive_2 1,
				fast_tac Lattice_cs 1];


(**** Symmetry ***************************************************************)

val symmetric_2 = spg "Symmetric(R) = (R^=R)";


(**** Vectors ****************************************************************)
				
val cmpl_predicate = rpg "-(bL) = -(bL)(L::('b,'b)rel)";
val cmpl_pred = pg "-(bL) = -(bL)L" [
				rtac incl_antisym 1,
				rtac luce1 1,
				rtac incl_refl 1,
				sstac L_is_vector 1,
				rtac L_def 1,
				REPEAT(EVERY[
				rtac (schroeder2 RS iffD2) 1,
				simp_tac Relation_ss 1,
				s_tac comp_assoc 1,
				sstac L_is_vector 1,
				rtac incl_refl 1])];

val cmpl_vector = pg "Vector(v) ==> Vector(-v)" [
				rewtac vector_def,
				etac subst 1,
				rtac sym 1,
				s_tac conv_equal 1,
				simp_tac Relation_ss 1,
				rtac cmpl_predicate 1];

val cmpl_vec = pg "-(Lv) = L-(Lv)" [
				rtac incl_antisym 1,
				rtac luce2 1,
				rtac incl_refl 1,
				sstac L_is_vector 1,
				rtac L_def 1,
				REPEAT(EVERY[
				rtac (schroeder1 RS iffD2) 1,
				simp_tac Relation_ss 1])];

val meet_vector = pg "[| Vector(v); Vector(w) |] ==> Vector(vw)" [
				rewtac vector_def,
				rtac incl_antisym 1,
				mon_tac comp_hdistr_meet1 1,
				asm_full_simp_tac Relation_ss 1,
				mon2_tac comp_isoton1 1,
				sstac I_def1 1,
				rtac incl_refl 1,
				rtac L_def 1];

val vector_1 = rpg "Q(RLv) = QRLv";
val predicate_1 = rpg "(pLR)Q = pLRQ";
val vector_1_cmpl=pg "Q(R-(Lv)) = QR-(Lv)" [
				sstac cmpl_vec 1,
				sstac cmpl_vec 1, (*2*back();*)
				rtac vector_1 1];

val vector_2 = rpg "(QLv)R = Q(Rv^L)";
val vector_2_cmpl = pg "(Q-(Lv))R = Q(R-(v^L))" [
				sstac cmpl_vec 1,
				sstac cmpl_pred 1,
				s_tac conv_L 1, (*2*back();*)
				s_tac conv_comp 1,
				s_tac conv_cmpl 1,
				rtac vector_2 1];

val vector_3 = rpg "(QbL)R = QRbL";
val vector_3_cmpl=pg "(Q-(bL))R = QR-(bL)" [
				sstac cmpl_pred 1,
				sstac cmpl_pred 1,
				rtac vector_3 1];

val vector_4 = rpg "(QRS)L = (QSR^)L";
val vector_5 = rpg "L(QRS) = L(RQ^S)";


(**** Coreflexivity **********************************************************)

val coreflexive_2 = spg "Coreflexive(e) --> e(IR) = eR";
					(* requires sym_lemma*)
val coreflexive_3 = spg "Coreflexive(e) --> (eR)^ = eR";  (* requires I_D *)
val coreflexive_4 = spg "Coreflexive(e) --> (eR^) = eR";
val coreflexive_5 = spg "Coreflexive(e) --> ILe = e";


(**** Points *****************************************************************)

val point_L = pg "Point(p) ==> pp^ = L" [
				rewtac point_def,
				safe_tac HOL_cs,
				simp_tac Relation_ss 1,
				dtac function_bump 1,
				etac ssubst 1,
				rewtac vector_def,
				s_tac conv_L 1,
				s_tac conv_comp 1,
				etac subst 1,
				rtac incl_refl 1];

val point_is_vec_atom=pg "Point(p) & Vector(v) --> (vp = (v=O | v=p))" [
				safe_tac HOL_cs,
				simp_tac Relation_ss 2,
				etac swap 1,
				rewrite_goals_tac [point_def, function_def],
				safe_tac HOL_cs,
				dtac univalent_equal 1,
				etac subst 1,
				safe_tac HOL_cs,
				rewtac vector_def,
				etac subst 1,
				dtac (tarski_rule RS iffD1) 1,
				etac ssubst 1, (*1*back();*)
				rtac L_def 1,
				rtac incl_refl 1];
val point_is_vec_atom_ = pg "[| Point(p); Vector(v); vp; vO |] ==> v=p" [
				cut_facts_tac [point_is_vec_atom] 1,
				fast_tac HOL_cs 1];

val point_O = pg "Point(p) & Point(q) --> ((p  q) = (pq^ = O))" [
				safe_tac HOL_cs,
				dtac point_L 2,
				cut_facts_tac [L_unequal_O] 2,
				asm_full_simp_tac HOL_ss 2,
				etac swap 2,
				etac sym 2,
				simp_tac Relation_ss 1,
				rtac (schroeder2 RS iffD2) 1,
				simp_tac Relation_ss 1,
				rotate_tac 1 1, double_prem_tac 1,
				rotate_tac 2 1, dtac point_is_vec_atom_ 1,
				rewtac point_def,
				etac conjE 1,
				etac meet_vector 1,
				etac conjE 1,
				etac cmpl_vector 1,
				rtac meet_lb1 1,
				etac swap 1,
				asm_full_simp_tac Relation_ss 1,
				etac swap2 1, (*2*back()*)
				s_tac incl_equivO 1,
				etac swap 1,
				rtac sym 1,
				rtac point_is_vec_atom_ 1,
				rewtac point_def,
				fast_tac HOL_cs 1,
				fast_tac HOL_cs 1,
				fast_tac HOL_cs 1,
				rewrite_goals_tac [function_def, total_def],
				safe_tac HOL_cs,
				cut_facts_tac [comp_O1, L_unequal_O] 1,
				etac swap 1, (*1*back();*)
				ra2tac trans 1,
				etac sym 1,
				rewtac vector_def,
				rotate_tac 2 1, etac ssubst 1,
				rewtac incl_def,
				atac 1];

val point_bump = pg "Point(p) --> p^qR = qpR" [
				rtac impI 1,
				rewtac point_def,
				rtac sym 1,
				rtac function_bump 1,
				fast_tac HOL_cs 1];

val point_is_atom = pg "Point(p) & Point(q) --> (Rp^q = (R=O | R=p^q))" [
				safe_tac HOL_cs,
				rtac O_def 2,
				rtac incl_refl 2,
				etac swap 1,
				ra2tac univalent_equal1 1,
				rewrite_goals_tac [point_def, function_def,
					vector_def, univalent_def, total_def],
				safe_tac HOL_cs,
				simp_tac (Relation_ss delsimps [comp_assoc]) 1,
				s_tac comp_assoc 1,
				mon_tac comp_isoton2 1,
				atac 2,
				sstac comp_assoc 1,
				mtac L_def 1,
				etac incl_intro1 1,
				rtac incl_intro1 1,
				rtac sym 1,
				rtac pred_type_equiv 1,
				rtac univalent_equal2 1,
				rewtac univalent_def,
				s_tac comp_assoc 1,
				etac ssubst 1, (*3*back();*)
				sstac conv_comp 1,
				simp_tac(HOL_ss addsimps [conv_L,conv_invol])1,
				eatac ssubst 1,
				ratac comp_isoton1 1,
				dtac (tarski_rule RS iffD1) 1,
				sstac comp_assoc 1,
				etac ssubst 1, (*4*back();*)
				rtac L_def 1];

